; EHCI capabilities registers

EHCI_HCS_NPORTS_MASK = 0xF
EHCI_HCS_PORT_POWER_CONTROL = 0x10

struc EHCI_CAP
{  .caplength       db ?
   .reserved        db ?
   .hciversion      dw ?
   .hcsparams       dd ?
   .hccparams       dd ?
   .hcsp_portroute  dq ?
}
virtual at 0
  ECHI_CAP EHCI_CAP
end virtual

; EHCI operational registers

EHCI_OP_RS = 1
EHCI_OP_PERIODIC_SCHED_EN = 1 shl 4
EHCI_OP_ASYNC_SCHED_EN = 1 shl 5
EHCI_OP_PERIODIC_SCHED_STAT = 1 shl 14
EHCI_OP_ASYNC_SCHED_STAT = 1 shl 15

struc EHCI_OP
{  .usbcmd                dd ?
   .usbsts                dd ?
   .usbintr               dd ?
   .frindex               dd ?
   .ctrldssegment         dd ?
   .periodiclistbase      dd ?
   .asynclistaddr         dd ?
   .reserved              dd dup(0x3f-0x1c)
   .configflag            dd ?
; start of PORTSC
   .portsc	              dd ?
}
virtual at 0
  EHCI_OP EHCI_OP
end virtual

; EHCI QT descriptor

EHCI_QTD_TERMINATE = 1
EHCI_QTD_HALTED = 1 shl 6
EHCI_QTD_ACTIVE = 1 shl 7
EHCI_QTD_PID = 3 shl 8
EHCI_QTD_CERR = 3 shl 10
EHCI_QTD_CPAGE = 7 shl 12
EHCI_QTD_TOTAL_LEN = (1 shl 15 - 1) shl 16
EHCI_QTD_TOGGLE = 1 shl 31
EHCI_QTD_TOGGLE_DATA0 = 0
EHCI_QTD_TOGGLE_DATA1 = 1 shl 31

struc EHCI_QTD
{  .next_qtd	          dd ?
   .alt_next_qtd          dd ?
   .token                dd ?
   .bufptrs              dd dup(5)
   .bufptrs64            dd dup(5)
}

virtual at 0
  EHCI_QTD EHCI_QTD
end virtual

; EHCI Queue Head

struc EHCI_QH
{  .horiz_link          dd ?
   .epchar              dd ?
   .epcaps              dd ?
   .current_td          dd ?
   ; EHCI_QTD overlay
}

virtual at 0
  EHCI_QH EHCI_QH
end virtual

struc EHCI
{  .cap        dd ? ; capabilities registers address
   .op         dd ? ; operational registers address
   .dummy_qh   dd ? ; "dummy qh" address
   .base       dd ? ; MMIO base address
   .bus        dd ? ; PCI bus
   .devfn      dd ? ; PCI device and function number
}

virtual at 0
  EHCI EHCI
end virtual

; Start EHCI controller
; ---------------------
; ehci_start(hci_t *controller)
proc ehci_start stdcall, hci:dword
       mov      esi, [hci]
	   mov      esi, [esi+EHCI.op]
	   mov      edi, [esi+EHCI_OP.usbcmd]
	   or       edi, EHCI_HC_OP_RS
	   mov      [esi+EHCI_OP.usbcmd], edi
	   ret
endp

; Stop EHCI controller
; --------------------
; ehci_stop(hci_t *controller)
proc ehci_stop stdcall, hci:dword
       mov      esi, [hci]
	   mov      esi, [esi+EHCI.op]
	   mov      edi, [esi+EHCI_OP.usbcmd]
	   and      edi, not EHCI_HC_OP_RS
	   mov      [esi+EHCI_OP.usbcmd], edi
	   ret
endp

; Reset EHCI controller
; ---------------------
; ehci_reset(hci_t *controller)
proc ehci_reset stdcall, hci:dword
		stdcall ehci_stop, [hci]
		; delay 10ms
        mov     esi, [hci]
		mov     esi, [esi+EHCI.op]
		mov     edi, [esi+EHCI_OP.usbsts]
		test    edi, (1 shl 12) ; test of HC_HALTED bit
		jnz     .end
		mov     edi, 2 ; HC_RESET bit
		mov     [esi+EHCI_OP.usbcmd], edi
		; now wait 100ms
		mov     ecx, 10
.loop:
        ; delay 10ms
		mov     edi, [esi+EHCI_OP.usbcmd]
		test    edi, 2
		jz      .end
		loop    .loop
.end:
		ret
endp

; Take ownership of EHCI controller
; ---------------------------------
; ehci_take(hci_t *controller)
proc ehci_take stdcall, hci:dword
endp

; Create EHCI structure
; ---------------------
; hci_t *ehci_create()
proc ehci_create stdcall
		; allocate memory for new controller HCI
		mov     esi, [hci]
		mov     esi, [esi+HCI.data]
		; allocate memory for new EHCI instance
		; and write pointer at HCI.data

; Set MMIO access

		stdcall PciRead32, [esi+EHCI.bus], [esi+EHCI.devfn], PCI_COMMAND
		or      eax, PCI_COMMAND_MEMORY
		and     eax, not PCI_COMMAND_IO
		stdcall PciWrite32, [esi+EHCI.bus], [esi+EHCI.devfn], PCI_COMMAND, eax
; Fill controller structure with EHCI function pointers
		;...
; Read MMIO base
		stdcall PciRead32, [esi+EHCI.bus], [esi+EHCI.devfn], USBBASE
		mov     [esi+EHCI.base], eax

		; Initialize 0 device
		stdcall usb_init_device, [ehci], 0
        mov     eax, [esi+EHCI.base]
		mov     [esi+EHCI.cap], eax
		mov     bl, [esi+EHCI.cap] ; read caplength
		add     eax, bl
		mov     [esi+EHCI.op], eax
		ret
endp

FLADJ = 0x61

; Do EHCI initialization
; ----------------------
; ehci_init(hci_t *controller)
proc ehci_init stdcall, hci:dword
		mov     esi, [hci]
		mov     esi, [esi+HCI.data] ; This is EHCI controller
		stdcall PciWrite8, [esi+EHCI.bus], [esi+EHCI.devfn], FLADJ, (60000 - 59488)/6
		stdcall ehci_start, [esi+HCI.data]
		mov     eax, 1
		mov     edi, [esi+EHCI.op]
		mov     [edi+EHCI_OP.configflag], eax ; take control over ports
; Enable USB interrupts (not implemented)
; Define periodic list (not yet implemented)
; Fill dummy qh
		; fill it by zeroes
		mov     edi, [esi+EHCI.dummy_qh]
		mov     [edi+EHCI_QH.horiz_link], EHCI_QH_TERMINATE
		mov     [edi+EHCI_QH.td + EHCI_QTD.next_qtd], EHCI_QTD_TERMINATE
		mov     [edi+EHCI_QH.td + EHCI_QTD.alt_next_qtd], EHCI_QTD_TERMINATE
; Fill periodic list by dummy QHs
; Enable periodic shedule
; Init root hub
        ret
endp

; Find closest USB2.0 hub
; -----------------------
; closest_usb2_hub(subdev_t *dev, int addr, int port)
proc closest_usb2_hub stdcall, dev:dword, addr:dword, port:dword
endp

; ---------------------------------------------------------------
;
;                        Root Hub stuff

struc EHCI_RH
{  .ports_num   dd ? ; number of ports
   .ports       dd ? ; pointer to ports address
   .devices     dd ? ; pointer to attached devices
}

virtual at 0
   EHCI_RH EHCI_RH
end virtual

; Initialization of Root Hub
; --------------------------
; ehci_rh_init(usbdev_t *dev)
proc ehci_rh_init stdcall, dev:dword
		mov esi, [dev]
		mov edi, ehci_rh_destroy ; function pointer
		mov [esi+USBDEV.destroy], edi
		mov edi, ehci_rh_poll ; function pointer
		mov [esi+USBDEV.poll], edi
		mov edi, [esi+USBDEV.controller]
		mov edi, [edi+EHCI.op]
endp

; Destroy Root Hub
; ----------------
; ehci_rh_destroy(usbdev_t *dev)
proc ehci_rh_destroy stdcall, dev:dword
endp

; Poll Root Hub
; -------------
; ehci_rh_poll(usbdev_t *dev)
proc ehci_rh_poll stdcall, dev:dword
endp

; Hand over port for companion controller
; ---------------------------------------
; ehci_rh_hand_over_port(usbdev_t *dev, u32 port)
proc ehci_rh_hand_over_port stdcall, dev:dword, port:dword
        mov esi, [dev]
		mov esi, [esi+USBDEV.data]
		mov esi, [esi+EHCI_RH.ports]
		lea esi, [esi+eax*4] ; calculate port offset
		mov eax, [port]
; Clear ConnectStatusChange before execution
		mov edi, [esi] ; read port register value
		or  edi, EHCI_P_CONN_STATUS_CHANGE
		mov [esi], edi ; write it back
; Lowspeed device. Hand over to companion controller
		mov edi, [esi] ; read port register value
		or  edi, EHCI_P_PORT_OWNER
		mov [esi], edi ; write it back
; Wait 100ms and testing for
; status change all the time
        mov ecx, 10
.loop:
		mov edi, [esi] ; read port register value
		test edi, EHCI_P_CONN_STATUS_CHANGE
		jz .changed
		; timeout 10ms
		loop .loop
.changed:
		cmp ecx, 0
		jz  .err ; timeout
		mov edi, [esi] ; read port register value
		; RW/C register, clear it by writing 1
		or  edi, EHCI_P_CONN_STATUS_CHANGE
		mov [esi], edi ; write it back
		ret
.err:
		mov eax, 1
		ret
endp

; Scan port for status changes
; ----------------------------
; ehci_rh_scanport(usbdev_t *dev, u32 port)
proc ehci_rh_scanport, dev:dword, port:dword
		mov esi, [dev]
		mov esi, [esi+USBDEV.data]
		mov eax, [port]
		mov edi, [esi+EHCI_RH.devices+eax*4]
		cmp edi, -1
		jz .handle
.detach:
		mov esi, [dev]
		mov edi, [esi+USBDEV.data]
		stdcall usb_detach_device, [esi+USBDEV.controller],
					[edi+EHCI_RH.devices+eax*4]
		mov [edi+EHCI_RH.devices+eax*4], -1
.handle:
; Check if PORT_OWNER bit set and clear it,
; to reset all stuff before
		mov esi, [dev]
		mov esi, [esi+USBDEV.data]
		lea esi, [esi+EHCI_RH.ports+eax*4]
		mov edi, [esi] ; read port status value
		test edi, EHCI_P_PORT_OWNER
		jz .cont
		or edi, not EHCI_P_PORT_OWNER
		mov [esi], edi ; write it back
.cont:
		mov edi, [esi] ; read port status again - do not remove!
		test edi, EHCI_P_CURR_CONN_STATUS
		jz .nothing
		; delay 100ms - USB2.0 specification, 9.1.2 chapter
		mov edi, [esi] ; read it again
		and edi, EHCI_P_LINE_STATUS
		cmp edi, EHCI_P_LINE_STATUS_LOWSPEED
		jnz .cont2
		sdtcall ehci_rh_hand_over_port, [dev], [port]
		cmp eax, 0
		jnz .return
.cont2:
; Deassert enable, assert reset.
; These must change automatically
		mov edi, [esi] ; read port status again
		and edi, not EHCI_P_PORT_ENABLE
		or edi, EHCI_P_PORT_RESET
		mov [esi], edi ; and write it back
		; Wait a bit while reset is active
		; delay 50ms - USB2.0 specification 7.1.7.5 chapter
; Deassert reset
		mov edi, [esi] ; read port status
		and edi, not EHCI_P_PORT_RESET
		mov [esi], edi ; and write it back
; Wait max 2ms (EHCI spec 2.3.9)
; for flag change to finish
		mov ecx, 20
.loop2:
		mov edi, [esi] ; read port status
		test edi, EHCI_P_PORT_RESET
		jnz .cont3
		;delay 100us
		loop .loop2
; Check if port reset successful
		mov edi, [esi] ; read port status
		test edi, EHCI_P_PORT_RESET
		jz .err
		; TODO: add check for full/high speed
		mov esi, [dev]
		stdcall usb_attach_device, [esi+USBDEV.controller], [esi+USBDEV.addr], [port], 2
		mov esi, [esi+USBDEV.data]
		mov ebx, [port]
		mov esi, [esi+EHCI_RH.devices+ebx*4]
		mov [esi], eax
.nothing:
		mov esi, [dev]
		mov eax, [port]
		mov esi, [esi+USBDEV.data]
; RW/C register, so clear it by writing 1
		lea esi, [esi+EHCI_RH.ports+eax*4]
		mov edi, [esi] ; read port status register
		or  edi, EHCI_P_CONN_STATUS_CHANGE
		mov [esi], edi
.return:
		ret
.err:
		; timeout error
		ret
endp

; ---------------------------------------------------------------
;
;                         Transactions

; +=====================================================+
; |###################### EHCI QH ######################|
; +-----------------------------------------------------+
; |  Horizontal Link Pointer                            |
; +-----------------------------------------------------+
; |          |          Maximum Packet Length           |
; |          |            Device Address                |
; |          |         Inactivate on Next Transaction   |
; |          |      Endpoint Number                     |
; |          |               Endpoint Speed             |
;                      .........
; +-----------------------------------------------------+
; |:::::::::::::::::::: QTD OVERLAY ::::::::::::::::::::|
;                      .........
; +-----------------------------------------------------+
; |:::::::::::::::::::: QH QTD CHAIN :::::::::::::::::::|
;                         .....
; +=====================================================+

; EHCI Bulk Transaction
; ---------------------
; int ehci_bulk(endpoint_t *ep, int size, void *data, int finalize)
proc ehci_bulk stdcall, ep:dword, size:dword, dat:dword, finalize:byte
endp

; EHCI Control Transaction
; ------------------------
; int ehci_control(usbdev_t *dev, int direction, int drlen,
;       void *devreq, int dalen, void *data)
proc ehci_control stdcall, dev:dword, dir:byte, drlen:dword,
        devreq:dword, dalen:dword, dat:dword
		locals
			hubaddr dd 0
			hubport dd 0
			non_hs_ctrl_ep db 0
		endl

		mov esi, [dev]
		mov eax, [esi+USBDEV.speed]
		cmp eax, 2 ; test, if this high-speed USB2.0 device
		jnz .continue
		stdcall closest_usb2_hub, [dev], [hubaddr], [hubport]
		cmp eax, 0
		jnz .ret1
		mov [non_hs_ctrl_ep], 1

.create_qtds:

.ret1: ; return 1;
		mov eax, 1
		ret
endp

; vim: ft=fasm

