; bRequest codes:
REQ_GET_STATUS = 0
REQ_CLEAR_FEATURE = 1
REQ_SET_FEATURE = 3
REQ_SET_ADDRESS = 5
REQ_GET_DESCRIPTOR = 6
REQ_SET_DESCRIPTOR = 7
REQ_GET_CONFIGURATION = 8
REQ_SET_CONFIGURATION = 9
REQ_GET_INTERFACE = 10
REQ_SET_INTERFACE = 11
REQ_SYNCH_FRAME = 12

; Device request structure
struc DEVREQ
{
  ; -- bmRequestType:
  ;      [7] - data direction
  ;      [5:6] - request type
  ;      [0:4] - request recipient
  .bmRequestType  db ?
  .bRequest       db ?
  .wValue         dw ?
  .wIndex         dw ?
  .wLength        dw ?
}
virtual at 0
  DEVREQ DEVREQ
end virtual

; USB endpoint structure
struc ENDPOINT
{ .device         dd ? ; pointer to parent device
  .endpoint       dd ? ; endpoint number
  .direction      dd ? ; direction
  .toggle         dd ?
  .maxpacketsz    dd ? ; Maximum packet size
  .type           dd ? ; Type of endpoint
}
virtual at 0
  ENDPOINT ENDPOINT
end virtual

ENDPOINTS_MAX = 32 ; Maximum possible number of device endpoints

; USB device structure
struc USBDEV
{ .controller     dd ? ; pointer to parent controller
  .endpoints      dd ? ; pointer to endpoints array (32)
  .endp_num       dd ? ; endpoints number
  .addr           dd ? ; USB address
  .hub            dd ? ; Hub, device is attached to
  .port           dd ? ; port where device is attached
  .speed          dd ? ; 1: lowspeed, 0: fullspeed, 2: highspeed
  .data           dd ? ; device-specific stuff
  .descriptor     dd ? ; USB descriptor pointer
  .configuration  dd ? ; USB configuration descriptor pointer
  .init           dd ? ; init() function pointer
  .destroy        dd ? ; destroy() function pointer
  .poll           dd ? ; poll() function pointer
}
virtual at 0
  USBDEV USBDEV
end virtual

TYPE_UHCI = 0
TYPE_OHCI = 1
TYPE_EHCI = 2
TYPE_XHCI = 3

; USB controller structure
struc HCI
{ .next           dd ? ; next HC in list
  .bus_addr       dd ? ; PCI address
  .base           dd ? ; base
  ;.highbase      dd ? ; unused in 32bit mode
  .type           dd ? ; HCI type: UHCI/OHCI/EHCI/XHCI
  .devices        dd ? ; pointer to attached devices (including hubs)
  .take           dd ? ; take() function - takes controller from BIOS/UEFI
  .start          dd ? ; start()
  .stop           dd ? ; stop()
  .reset          dd ? ; reset()
  .shutdown       dd ? ; shutdown()
  .bulk           dd ? ; bulk() - bulk transaction initiator
  .control        dd ? ; control() - control transaction initiator
  .create_intr_q  dd ? ; create_intr_queue()
  .destroy_intr_q dd ? ; destroy_intr_queue()
  .poll_intr_q    dd ? ; poll_intr_queue()
  .instance       dd ? ; controller specific stuff
}
virtual at 0
  HCI HCI
end virtual

; -----------------------------------------------------------------
;                      USB descriptors:

; Hub descriptor
struc HUB_DESCRIPTOR
{ .bDescLength             db ?
  .bDescriptorType         db ?
  .bNbrPorts               db ?
; -- wHubCharacteristics:
;      [0:1] -  logicalPowerSwitchingMode
;       [2]  - isCompoundDevice
;      [3:4] - overcurrentProtectionMode
;      [5:6] - ttThinkTime
;       [7]  - arePortIndicatorsSupported
;      [8:15]  - RESERVED
  .wHubCharacteristics     dw ?
  .bPowerOn2PwrGood        db ?
  .bHubContrCurrent        db ?
  .DeviceRemovable         db ?
}
virtual at 0
  HUB_DESCRIPTOR HUB_DESCRIPTOR
end virtual

; Device descriptor
struc DEV_DESCRIPTOR
{ .bLength                 db ?
  .bDescriptorType         db ?
  .bcdUSB                  dw ?
  .bDeviceClass            db ?
  .bDeviceSubClass         db ?
  .bDeviceProtocol         db ?
  .bMaxPacketSize0         db ?
  .idVendor                dw ?
  .idProduct               dw ?
  .bcdDevice               dw ?
  .iManufacturer           db ?
  .iProduct                db ?
  .iSerialNumber           db ?
  .bNumConfigurations      db ?
}
virtual at 0
  DEV_DESCRIPTOR DEV_DESCRIPTOR
end virtual

; Configuration descriptor
struc CONF_DESCRIPTOR
{ .bLength                 db ?
  .bDescriptorType         db ?
  .wTotalLength            dw ?
  .bNumInterfaces          db ?
  .bConfigurationValue     db ?
  .iConfiguration          db ?
  .bmAttributes            db ?
  .bMaxPower               db ?
}
virtual at 0
  CONF_DESCRIPTOR CONF_DESCRIPTOR
end virtual

; Interface descriptor
struc INTERF_DESCRIPTOR
{ .bLength                 db ?
  .bDescriptorType         db ?
  .bInterfaceNumber        db ?
  .bAlternateSetting       db ?
  .bNumEndpoints           db ?
  .bInterfaceClass         db ?
  .bInterfaceSubClass      db ?
  .bInterfaceProtocol      db ?
  .iInterface              db ?
}
virtual at 0
  INTERF_DESCRIPTOR INTERF_DESCRIPTOR
end virtual

; Endpoint descriptor
struc ENDP_DESCRIPTOR
{ .bLength                 db ?
  .bDescriptorType         db ?
  .bEndpointAnddress       db ?
  .bmAttributes            db ?
  .wMaxPacketSize          dw ?
  .bInterval               db ?
}
virtual at 0
  ENDP_DESCRIPTOR ENDP_DESCRIPTOR
end virtual

; HID descriptor
struc HID_DESCRIPTOR
{ .bLength                 db ?
  .bDescriptorType         db ?
  .bcdHID                  dw ?
  .bCountryCode            db ?
  .bNumDescriptors         db ?
  .bReportDescriptorType   db ?
  .wReportDescriptorLength dw ?
}
virtual at 0
  HID_DESCRIPTOR HID_DESCRIPTOR
end virtual

; -------------------------------------------------

proc new_controller stdcall, list:dword
endp

proc detach_controller stdcall, list:dword, hci:dword
endp

proc usb_exit stdcall, list:dword
endp

; Polls all hubs on all USB controllers, to find
; out about device changes
; ------------------------
; usb_poll(hci_t *list)
proc usb_poll stdcall, list:dword
endp


proc init_device_entry stdcall, hci:dword, i:dword
endp

; set_feature(usbdev_t *dev, int ep, int feature, int rtype)
proc set_feature stdcall, dev:dword, ep:dword, ft:dword, rtype:dword
endp

; get_status(usbdev_t *dev, int intf, int rtype, int len, void *data)
proc get_status stdcall, dev:dword, intf:dword, rtype:dword, len:dword, dat:dword

endp

; void *get_descriptor(usbdev_t *dev, int reqtype, int descType,
;        int descIdx, int langId)
proc get_descriptor stdcall, dev:dword, rtype:byte, dtype:dword, idx:dword, lang:dword
endp

; vim: ft=fasm

